import csv
import logging
import os
import sqlite3
import subprocess
import zipfile

"""
This script automates the process of downloading backups from the MyWealth platform, extracting the data, and loading it into a SQLite database.

Scheduling via Linux Crontab:

To automate the running of this script, you can use cron on a Linux system.
1. First ensure that the following environment variables are correctly set for the user that will be executing this script
    MWA_ENVIRONMENT: The environment to connect to (dev, sandbox, prod)
    MWA_TENANT: The tenant to connect to
    MWA_USERNAME: The username to use for login
    MWA_PASSWORD: The password to use for login

2. Open the crontab file:

## crontab -e

This will open the crontab file in your default editor.

3. Add an entry to run the script at a scheduled time. For example, to run the script at 3:00 AM every day:

0 3 * * * /usr/bin/python3 /path/to/your/mywealth_backup.py

This runs the script every day at 3:00 AM. Replace /path/to/your/mywealth_backup.py with the actual path to this script.

4. Save and exit the crontab editor. The cron job will now execute at the scheduled time.

5. Optionally make sure that the script is executable if you manually downloaded it:

## chmod +x /path/to/your/mywealth_backup.py

6. If you're running the script in an environment where a virtual environment is used, you may need to specify the Python binary from the virtual environment in the cron command:

0 3 * * * /path/to/venv/bin/python /path/to/your/mywealth_backup.py

This will ensure that all required dependencies are used correctly when the script is executed from the cron job.

"""


def setup_logging():
    # Configure logging
    logging.basicConfig(filename='backup.log', level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(message)s')
    # Create a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
    # Add the console handler to the root logger
    logging.getLogger().addHandler(console_handler)


def process_files(directory, database):
    # Connect to SQLite database (or create it if it doesn't exist)
    try:
        conn = sqlite3.connect(database)
        cursor = conn.cursor()
        logging.info(f"Connected to SQLite database: {database}")
    except sqlite3.Error as e:
        logging.error(f"Error connecting to database: {e}")
        raise

    # Iterate through all files in the directory
    for file_name in os.listdir(directory):
        if file_name.endswith('.csv.zip'):
            try:
                # Construct the full file path
                file_path = os.path.join(directory, file_name)

                # Extract the CSV file from the zip
                with zipfile.ZipFile(file_path, 'r') as zip_ref:
                    zip_ref.extractall(directory)
                    logging.info(f"Extracted file from zip: {file_name}")

                # Get the CSV file name (assumes one file per zip)
                csv_file_name = file_name.replace('.zip', '')
                csv_file_path = os.path.join(directory, csv_file_name)

                # Load the CSV into SQLite
                load_csv_into_sqlite(csv_file_path, cursor)

                # Delete the extracted CSV file to clean up
                os.remove(csv_file_path)
                logging.info(f"Deleted extracted CSV file: {csv_file_name}")
            except Exception as e:
                logging.error(f"Error processing {file_name}: {e}")
                continue  # Skip this file and move to the next

    # Commit changes and close the connection
    try:
        conn.commit()
        conn.close()
        logging.info("Database commit and connection closed successfully.")
    except sqlite3.Error as e:
        logging.error(f"Error committing changes or closing database: {e}")
        raise


def load_csv_into_sqlite(csv_file, cursor):
    # Extract table name from the file name
    table_name = os.path.splitext(os.path.basename(csv_file))[0]

    # "order" is a reserved keyword in SQLite, so we will use "orders" instead if that is passed in
    if table_name == "order":
        table_name = "orders"

    try:
        with open(csv_file, 'r') as f:
            reader = csv.reader(f)
            headers = next(reader)  # Get the first row as headers

            # Create table if it doesn't exist
            create_table_query = f"""
            CREATE TABLE IF NOT EXISTS {table_name} (
                {headers[0]} BIGINT PRIMARY KEY,
                {', '.join([f'{header} TEXT' for header in headers[1:]])}
            )
            """
            cursor.execute(create_table_query)

            # Prepare the upsert query
            placeholders = ', '.join(['?' for _ in headers])
            update_placeholders = ', '.join([f'{header}=excluded.{header}' for header in headers])
            upsert_query = f"""
            INSERT INTO {table_name} ({', '.join(headers)})
            VALUES ({placeholders})
            ON CONFLICT({headers[0]}) DO UPDATE SET {update_placeholders}
            """

            # Insert data into the table
            for row in reader:
                cursor.execute(upsert_query, row)
        logging.info(f"Loaded data into {table_name} table successfully.")
    except Exception as e:
        logging.error(f"Error processing data for {table_name}: {e}")
        raise


def run_cli_login_command(environment, tenant, username, password, exe_path="mwa"):
    try:
        # Run the login command
        command = f"{exe_path} --environment {environment} --tenant {tenant} login --username {username} --password {password}"
        result = subprocess.run(command, shell=True, check=True, capture_output=True, text=True)
        logging.info(f"Login completed successfully: {result.stdout}")
    except subprocess.CalledProcessError as ex:
        logging.error(f"Error logging in: {ex.stderr}")
        raise


def run_cli_backup_command(environment, tenant, backup_type, exe_path="mwa"):
    try:
        # Run the backup command
        command = f"{exe_path} --environment {environment} --tenant {tenant} backup --backup_type {backup_type}"
        result = subprocess.run(command, shell=True, check=True, capture_output=True, text=True)
        logging.info(f"Backup completed successfully: {result.stdout}")
    except subprocess.CalledProcessError as ex:
        logging.error(f"Error running backup command: {ex.stderr}")
        raise


def get_latest_backup_directory(base_directory):
    # Find the latest folder inside base_directory
    try:
        directories = [d for d in os.listdir(base_directory) if os.path.isdir(os.path.join(base_directory, d))]
        latest_dir = max(directories, key=lambda d: os.path.getmtime(os.path.join(base_directory, d)))
        latest_dir_path = os.path.join(base_directory, latest_dir)
        logging.info(f"Latest backup folder found: {latest_dir_path}")
        return latest_dir_path
    except Exception as ex:
        logging.error(f"Error finding the latest backup folder: {ex}")
        raise


def validate_install_cli():
    # Check if the mwa CLI is installed and download it if not found in the current directory
    os_linux_x86_64_url = "https://mywealth-cli-download.s3.us-east-2.amazonaws.com/linux/x86_64/mwa"
    os_linux_arm64_url = "https://mywealth-cli-download.s3.us-east-2.amazonaws.com/linux/aarch64/mwa"
    os_mac_x86_64_url = "https://mywealth-cli-download.s3.us-east-2.amazonaws.com/apple/x86_64/mwa"
    os_mac_arm64_url = "https://mywealth-cli-download.s3.us-east-2.amazonaws.com/apple/aarch64/mwa"
    os_windows_x86_64_url = "https://mywealth-cli-download.s3.us-east-2.amazonaws.com/windows/x86_64/mwa.exe"

    # Get the current operating system
    os_name = os.uname().sysname.lower()
    os_machine = os.uname().machine.lower()
    exe_file_name = "mwa"
    if os_name == "linux":
        if os_machine == "x86_64":
            exe_url = os_linux_x86_64_url
        elif os_machine == "aarch64":
            exe_url = os_linux_arm64_url
        else:
            logging.error("Unsupported Linux architecture.")
            exit(1)
    elif os_name == "darwin":
        if os_machine == "x86_64":
            exe_url = os_mac_x86_64_url
        elif os_machine == "arm64":
            exe_url = os_mac_arm64_url
        else:
            logging.error("Unsupported macOS architecture.")
            exit(1)
    elif os_name == "windows":
        if os_machine == "x86_64":
            exe_url = os_windows_x86_64_url
            exe_file_name = "mwa.exe"
        else:
            logging.error("Unsupported Windows architecture.")
            exit(1)
    else:
        logging.error("Unsupported operating system.")
        exit(1)

    # Check if the mwa CLI is installed and download it if not found in the current directory
    if not os.path.exists(exe_file_name):
        logging.info(f"Downloading the mwa CLI... for the first time, platform {os_name} & download url {exe_url}")
        try:
            # Download the mwa CLI executable
            subprocess.run(f"curl -LJO {exe_url}", shell=True, check=True)

            # Make the downloaded file executable
            os.chmod(exe_file_name, 0o755)

            logging.info("Download complete!")
        except subprocess.CalledProcessError as ex:
            logging.error(f"Error downloading the mwa CLI: {ex.stderr}")
            exit(1)
    return exe_file_name


if __name__ == "__main__":
    # Setup logging
    setup_logging()

    # Read the configuration from environment variables or hardcode them here
    _environment = os.getenv("MWA_ENVIRONMENT", "dev")  # Replace with your environment (dev, sandbox, prod)
    _tenant = os.getenv("MWA_TENANT", "default")  # Replace with your tenant
    _username = os.getenv("MWA_USERNAME", "")  # Replace with your username which is typically your email
    _password = os.getenv("MWA_PASSWORD", "")  # Replace with your password
    _backup_type = os.getenv("MWA_BACKUP_TYPE", "full")  # Replace with your backup type (full, incremental)

    # Validate that the backup type is either 'full' or 'incremental'
    if _backup_type not in ["full", "incremental"]:
        logging.error("Invalid backup type. Please specify either 'full' or 'incremental'.")
        exit(1)

    # Validate the environment
    if _environment not in ["dev", "sandbox", "prod"]:
        logging.error("Invalid environment. Please specify either 'dev', 'sandbox', or 'prod'.")
        exit(1)

    # Validate that the username and password are not empty
    if not _username or not _password:
        logging.error("Username and password are required.")
        exit(1)

    # Validate that the mwa CLI is installed and get the executable name
    exe_name = validate_install_cli()

    # Update the PATH environment variable to include the current directory
    os.environ["PATH"] += os.pathsep + os.getcwd()

    # Print the configuration for logging - environment, tenant, backup type
    logging.info(f"Environment: {_environment}")
    logging.info(f"Tenant: {_tenant}")
    logging.info(f"Backup Type: {_backup_type}")

    try:
        # Step 1: Login
        run_cli_login_command(_environment, _tenant, _username, _password, exe_name)

        # Step 2: Run backup
        run_cli_backup_command(_environment, _tenant, _backup_type, exe_name)

        # Step 3: Find the latest backup folder
        backup_base_dir = f"data/{_tenant}"
        latest_backup_dir = get_latest_backup_directory(
            os.path.join(backup_base_dir, _environment, "bak", _backup_type))

        # Step 4: Process files in the latest backup folder
        process_files(latest_backup_dir, "mywealth_backup.db")  # Your SQLite database file name
        logging.info("Backup process has completed!!")
    except Exception as e:
        logging.error(f"Error during script execution: {e}")
        exit(1)
